"""Main Pytest Plugin Scaffolding."""
import logging
import os
import time

from pytest_f3ts import schemas
from pytest_f3ts.runner_client import F3TSASCIIAPI, F3TSBackendAPI

logger = logging.getLogger(__name__)


def safe_format(value, default=" ", width=15):
    """Return a formatted string or a default value if None."""
    if value is None:
        value = default
    value = str(value)

    if len(value) > width:
        value = value[: width - 3] + "..."

    return f"{value:<{width}}"


def pytest_addoption(parser):
    """Add F3TS Plugin related options."""
    group = parser.getgroup("f3ts", "f3ts test plugin")

    group.addoption(
        "--test_name",
        action="store",
        help="Test name, used for creating the log filename",
    )
    group.addoption(
        "--run_id",
        action="store",
        default="DEV",
        help="Database Run ID to send results to",
    )
    group.addoption(
        "--plan_id",
        action="store",
        default="DEV",
        help="Database Plan ID to grab configuration information from",
    )
    group.addoption(
        "--slot_id",
        action="store",
        default="1",
        help="Slot ID of fixture that ran the test",
    )
    group.addoption(
        "--frontend_client_id",
        action="store",
        default="backend",
        help="Runner websocket client ID to connect to",
    )
    group.addoption(
        "--serial_number",
        action="store",
        help="Serial Number of the device being tested",
    )
    group.addoption(
        "--f3ts_config",
        action="store",
        default="config.yml",
        help="F3TS Configuration File",
    )
    group.addoption(
        "--backend_url",
        action="store",
        help="Backend API",
    )
    group.addoption(
        "--fixture_url",
        action="store",
        default="http://localhost:8886",
        help="Fixture API",
    )
    group.addoption("--teardown", action="store_true")
    group.addoption(
        "--skip-summary",
        action="store_true",
        help="Skip FixturFab Test Runner Report Summary",
    )


def pytest_configure(config):
    """Add markers and other configuration information for the F3TS plugin."""
    if not config.option.teardown:
        p = F3TSReporter(config)
        config.pluginmanager.register(p, "f3ts_reporter")


class F3TSReporter:
    """F3TS Test Runner Reporter Plugin.

    This plugin implements all backend communication with the
    F3TS Test Runner application.
    """

    start_time = None

    def __init__(self, config):
        self.config = config
        self.api = None
        self.run_id = self.config.option.run_id
        self.slot_id = self.config.option.slot_id
        self.frontend_client_id = self.config.option.frontend_client_id

        try:
            self.api = F3TSBackendAPI(
                api_url=self.config.option.backend_url,
                run_id=self.run_id,
                slot_id=self.slot_id,
                frontend_client_id=self.frontend_client_id,
            )
            self.operator_id = self.api.operator_id

        except Exception as e:
            logger.warning(
                "WARNING: Unable to connect to backend API, "
                f"not sending results to backend!\n{e}"
            )
            self.run_id = 0
            self.operator_id = 0
            self.api = F3TSASCIIAPI(slot_id=self.slot_id)

    def pytest_collection_modifyitems(self, items):
        """Pre-processing after collection has been called."""
        self.start_time = time.time()
        terminalreporter = self.config.pluginmanager.getplugin("terminalreporter")

        tests = [i.originalname for i in items]
        terminalreporter.ensure_newline()
        terminalreporter.line(f"tests collected: {tests}")

    def pytest_runtest_logreport(self, report):
        """Send the unit test report to the Backend.

        Create a TestReport model from the `pytest` `report` object
        and send the results to the backend webserver. Send error
        notification after failed test.
        """
        terminalreporter = self.config.pluginmanager.getplugin("terminalreporter")

        # Only send results for "call" actions
        if report.when == "call":
            # Create base TestResult
            result = schemas.ResultCreate(
                test_name=report.head_line,
                operator_id=self.operator_id,
                start=report.start,
                stop=report.stop,
                duration=report.duration,
                passed=report.passed,
            )

            # Add user keys
            user_properties_dict = dict()
            for user_property in report.user_properties:
                user_properties_dict[user_property[0]] = user_property[1]

            search_props = [
                "test_id",
                "description",
                "error_code",
                "error_msg",
                "min_limit",
                "max_limit",
                "meas",
            ]
            if "test_config" in user_properties_dict.keys():
                config_dict = dict(user_properties_dict["test_config"])
                for k, v in config_dict.items():
                    if k in search_props:
                        setattr(result, k, v)

            for k, v in user_properties_dict.items():
                if k in search_props:
                    setattr(result, k, v)

            terminalreporter.ensure_newline()
            terminalreporter.section(f"Test ID: {result.test_id}", sep="-")

            terminalreporter.line(str(result))
            self.api.send_test_result(result)

            # Send error message notification is test failed:
            if not result.passed:
                error_message = "Unknown Error Occured"
                error_code = ""

                if result.error_msg:
                    error_message = result.error_msg

                if result.error_code:
                    error_code = result.error_code

                    error_notif = schemas.Notif(
                        title=f"ERROR {error_code}: {result.test_name}",
                        message=error_message,
                        okButtonText="OK",
                    )
                    print(error_notif)
                    self.api.send_user_notif(error_notif)

    def pytest_sessionfinish(self, session):
        """Send the test finished message to the Backend."""

    def pytest_terminal_summary(self, terminalreporter, exitstatus, config):
        """Send collected run information to database.

        Updates information such as duration, serial number, fixture id, etc.
        """

        duration = time.time() - self.start_time
        run = schemas.RunUpdate(
            slot_id=config.option.slot_id,
            operator_id=self.operator_id,
            duration=duration,
            passed=(exitstatus == 0),
            num_failed=(len(terminalreporter.stats.get("failed", []))),
            num_passed=(len(terminalreporter.stats.get("passed", []))),
            num_skipped=(len(terminalreporter.stats.get("skipped", []))),
        )

        terminalreporter.ensure_newline()
        terminalreporter.section("Fixturfab Test Runner (Updating Run...)", sep="-")
        terminalreporter.line(str(run))
        self.api.update_run_data(run)

        # Generate Test Report Summary if no backend API is connected:
        if self.api.api_url is None and not config.option.skip_summary:
            terminalreporter.ensure_newline()
            terminalreporter.section(" FixturFab Test Runner Summary ", bold=True)
            terminalreporter.ensure_newline()
            width = os.get_terminal_size().columns - 71
            if width < 60:
                name_width = 25
                desc_width = 35
            else:
                name_width = int((width * 3) / 8)
                desc_width = width - name_width

            # Header
            header = f"{'Test ID ':<9} {'Test Name':<{name_width}} {'Description':<{desc_width}} {'Min Limit':<11} {'Measurement':<13} {'Max Limit':<11} {'Passed':<8} {'Duration (s)':<10}"
            terminalreporter.write("-" * len(header) + "\n")
            terminalreporter.write(header + "\n")
            terminalreporter.write("-" * len(header) + "\n")

            # Iterate over all collected test reports
            for result in self.api.results:
                terminalreporter.write(
                    f"{safe_format(result.test_id, width=9)} "
                    f"{safe_format(result.test_name, width=name_width)} "
                    f"{safe_format(result.description, width=desc_width)} "
                    f"{safe_format(result.min_limit, width=11)} "
                    f"{safe_format(result.meas, width=13)} "
                    f"{safe_format(result.max_limit, width=11)} "
                    f"{safe_format(str(result.passed), default='N/A', width=8)} "
                    f"{safe_format(round(result.duration, 5), default='N/A', width=10)}\n"
                )

            terminalreporter.write("-" * len(header) + "\n")



"""Test Runner API Client."""
import os

import requests
from fastapi import HTTPException
from requests import JSONDecodeError
from tenacity import retry, stop_after_attempt

from . import schemas


class F3TSBackendAPI:
    """F3TS Backend Webservice API.

    API Client for communicating with the FixturFab Test Runner Backend. This
    class is used to send test results and other information to the Backend.

    TODO: Implement API Key Authentication
    """

    def __init__(
        self,
        api_url: str = "https://backend:8888/api/v1",
        slot_id: str = "1",
        run_id: int = None,
        plan_id: int = None,
        frontend_client_id: str = "frontend",
        retries=3,
    ):
        """Initialize the F3TS Backend API.

        Store the API URL and number of request retries. A request will be
        retried for the provided number of retries. If it continues to fail,
        an exception will be raised.
        """
        self.api_url = api_url
        self.run_id = run_id
        self.plan_id = plan_id
        self.slot_id = slot_id
        self.frontend_client_id = frontend_client_id
        self.frontend_group_id = "frontend_" + self.slot_id
        self.retries = retries
        self.headers = {
            "accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
        }
        # self.access_token = self.authenticate(username, password)
        self.operator_id = None

        self.headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
        }

        if run_id:
            self.backend_client_id = str(run_id)
            self.operator_id = self.get_operator_id()

        else:
            self.backend_client_id = "prestart"

    # def authenticate(self, username, password):
    #     data = {"username": username, "password": password}
    #     response = requests.post(
    #         f"{self.api_url}/login/access-token", headers=self.headers, data=data
    #     )
    #     if response.status_code == 200:
    #         access_token = schemas.Token.parse_obj(response.json()).access_token
    #         return access_token
    #     else:
    #         raise HTTPException(
    #             response.status_code, "Unable to authenticate login credentials."
    #         )

    @retry(stop=stop_after_attempt(3))
    def get_config(self, slot_id=None):
        """Return a configuration object."""

        response = requests.get(
            f"{self.api_url}/plans/{self.plan_id}",
            headers=self.headers,
            verify=False,
        )

        if response.status_code == 200:
            plan = schemas.Plan.parse_obj(response.json())

            if slot_id:  # Parse into config for specified fixture:
                my_fixture_settings = schemas.FixtureSettings(slot_id=slot_id)
                for fixture_config in plan.fixture_settings.values():
                    if slot_id == fixture_config.slot_id:
                        my_fixture_settings = fixture_config

                f3ts_config_obj = schemas.PytestTestPlanConfig(
                    test_name=plan.name,
                    test_cases=plan.limits_settings.test_cases,
                    runner_settings=plan.runner_settings,
                    fixture_settings=my_fixture_settings,
                )

            else:
                f3ts_config_obj = schemas.TestPlanConfig(
                    test_name=plan.name,
                    test_cases=plan.limits_settings.test_cases,
                    runner_settings=plan.runner_settings,
                    fixture_settings=plan.fixture_settings,
                )

            return f3ts_config_obj

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to read test plan in backend: {data}",
        )

    @retry(stop=stop_after_attempt(3))
    def get_operator_id(self):
        """Get operator id from run information stored in database"""
        response = requests.get(
            f"{self.api_url}/runs/{self.run_id}",
            headers=self.headers,
            verify=False,
        )

        if response.status_code == 200:
            run = schemas.Run.parse_obj(response.json())
            self.operator_id = run.operator_id
            return run.operator_id

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to update test run in backend: {data}",
        )

    @retry(stop=stop_after_attempt(3))
    def update_run_data(self, run: schemas.RunUpdate):
        """Update run data with run specific information

        Add information such as log filename, serial number, etc.
        """
        response = requests.post(
            f"{self.api_url}/runs/{self.run_id}",
            data=run.json(exclude_unset=True),
            headers=self.headers,
            verify=False,
        )

        if response.status_code == 200:
            return True

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to update test run in backend: {data}",
        )

    @retry(stop=stop_after_attempt(3))
    def save_logfile(self, attachment: schemas.AttachmentCreate):
        """Save logfile from local shared directory."""
        response = requests.post(
            f"{self.api_url}/runs/{self.run_id}/attachment/",
            params={"run_id": self.run_id},
            data=attachment.json(exclude_unset=True),
            headers=self.headers,
            verify=False,
        )

        if response.status_code == 200:
            return True

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to save log file for run: {data}",
        )

    @retry(stop=stop_after_attempt(3))
    def upload_logfile(self, attachment: schemas.AttachmentCreate):
        """Upload logfile over request."""

        response = requests.post(
            f"{self.api_url}/runs/{self.run_id}/attachment/upload",
            params={
                "run_id": self.run_id,
                "outfile_dir": os.path.dirname(attachment.filepath),
            },
            files={"file": open(attachment.filepath, "rb")},
            headers={"accept": "application/json"},
            verify=False,
        )

        if response.status_code == 200:
            return True

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to upload log file to run: {data}",
        )

    @retry(stop=stop_after_attempt(3))
    def send_websocket_msg(self, message: schemas.SocketMsg):
        """Send a `pytest` test case result to the Backend."""
        response = requests.post(
            f"{self.api_url}/utils/runner_socket/send",
            data=message.json(),
            headers=self.headers,
            verify=False,
        )

        if response.status_code == 200:
            return True

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to send websocket msg to frontend: {data}",
        )

    @retry(stop=stop_after_attempt(3))
    def send_test_result(self, result: schemas.ResultCreate):
        """Send a `pytest` test case result to the Backend."""
        # Send via websocket to frontend
        message = schemas.SocketMsg(
            type="result",
            toClient=self.frontend_group_id,
            fromClient=self.backend_client_id,
            data=result,
        )

        self.send_websocket_msg(message)

        response = requests.post(
            f"{self.api_url}/runs/{self.run_id}/result/",
            data=result.json(),
            headers=self.headers,
            verify=False,
        )

        if response.status_code == 200:
            return True

        try:
            data = response.json()
        except JSONDecodeError:
            data = response.text

        raise HTTPException(
            response.status_code,
            f"Unable to send test result to backend: {data}",
        )

    def send_user_notif(self, notif: schemas.Notif):
        """Send one-way user dialog message to the Backend."""

        message = schemas.SocketMsg(
            type="notif",
            toClient=self.frontend_group_id,
            fromClient=self.backend_client_id,
            data=notif,
        )

        self.send_websocket_msg(message)

    def send_banner_update(self, banner: schemas.StatusBanner):
        """Send set banner message to the Backend."""

        message = schemas.SocketMsg(
            type="banner",
            toClient=self.frontend_group_id,
            fromClient=self.backend_client_id,
            data=banner,
        )

        self.send_websocket_msg(message)

    def send_serial_number(self, serial_number):
        """Send serial number to the Backend."""

        run = schemas.RunUpdate(serial_number=serial_number)
        self.update_run_data(run)

        message = schemas.SocketMsg(
            type="serial",
            toClient=self.frontend_group_id,
            fromClient=self.backend_client_id,
            data={"serial_number": serial_number},
        )

        self.send_websocket_msg(message)


class F3TSASCIIAPI:
    """F3TS Backend Webservice API.

    API Client for communicating with the FixturFab Test Runner Backend. This
    class is used to send test results and other information to the Backend.

    TODO: Implement API Key Authentication
    """

    def __init__(
        self,
        slot_id: str = "1",
    ):
        """Initialize the F3TS ASCII API.

        Store the API URL and number of request retries. A request will be
        retried for the provided number of retries. If it continues to fail,
        an exception will be raised.
        """
        self.api_url = None
        self.slot_id = slot_id
        self.operator_id = 0
        self.run_id = 0
        self.results = []
        self.notifs = []

    # def authenticate(self, username, password):
    #     data = {"username": username, "password": password}
    #     response = requests.post(
    #         f"{self.api_url}/login/access-token", headers=self.headers, data=data
    #     )
    #     if response.status_code == 200:
    #         access_token = schemas.Token.parse_obj(response.json()).access_token
    #         return access_token
    #     else:
    #         raise HTTPException(
    #             response.status_code, "Unable to authenticate login credentials."
    #         )

    @retry(stop=stop_after_attempt(3))
    def get_config(self, slot_id=None):
        """Return a configuration object."""
        self.slot_id = slot_id

    @retry(stop=stop_after_attempt(3))
    def update_run_data(self, run: schemas.RunUpdate):
        """Update run data with run specific information

        Add information such as log filename, serial number, etc.
        """
        self.run = run

    @retry(stop=stop_after_attempt(3))
    def upload_logfile(self, attachment: schemas.AttachmentCreate):
        """Upload logfile."""
        self.attachment = attachment

    @retry(stop=stop_after_attempt(3))
    def send_test_result(self, result: schemas.ResultCreate):
        """Send a `pytest` test case result to the Backend."""
        self.results += [result]

    def send_user_notif(self, notif: schemas.Notif):
        """Send one-way user dialog message to the Backend."""
        self.notifs += [notif]

    def send_banner_update(self, banner: schemas.StatusBanner):
        """Send set banner message to the Backend."""
        self.banner = banner

    def send_serial_number(self, serial_number):
        """Send serial number to the Backend."""
        self.serial_number = serial_number



from pydantic import BaseModel


class Msg(BaseModel):
    msg: str


class Dialog(BaseModel):
    title: str = "INFO:"
    message: str = " Enter dialog message here"
    okButtonText: str = "OK"
    cancelButtonText: str = "Cancel"
    inputType: str = None
    defaultText: str = ""


class Notif(BaseModel):
    title: str = "INFO:"
    message: str = " Enter notification message here"
    okButtonText: str = "OK"


class DialogResponse(BaseModel):
    inputText: str = ""
    okClicked: bool = False
    cancelClicked: bool = False


class StatusBanner(BaseModel):
    status: str = None
    color: str = None


class SocketMsg(BaseModel):
    type: str
    toClient: str
    fromClient: str
    data: dict = None



"""Test Run Schema"""
from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from .limits import Limits
from .result import Result


class RunBase(BaseModel):
    """Test Run Base schema."""

    operator_id: Optional[int] = None
    serial_number: Optional[str] = None
    slot_id: Optional[str] = None
    api_url: Optional[str] = None
    test_plan_version: Optional[str] = None
    start: Optional[str] = None
    duration: Optional[float] = None
    passed: Optional[bool] = None
    log_file: Optional[str] = None
    num_failed: Optional[int] = None
    num_passed: Optional[int] = None
    num_skipped: Optional[int] = None


class RunCreate(RunBase):
    """Test Run Create schema."""

    operator_id: int
    slot_id: str
    api_url: str
    start: str = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")


class RunUpdate(RunBase):
    """Test Run Update schema."""

    pass


class RunInDBBase(RunBase):
    """Test Run Database base schema."""

    id: int
    plan_id: int
    limits_id: Optional[int]
    results_id: Optional[int] = None
    results: Optional[List[Result]] = None
    limits: Optional[Limits] = None

    class Config:
        orm_mode = True


class Run(RunInDBBase):
    """Test Run schema."""

    pass


class RunInDB(RunInDBBase):
    pass



"""Git/Balena Commit Schema"""
from typing import Optional

from pydantic import BaseModel


class CommitBase(BaseModel):
    balena_id: Optional[int] = None
    balena_commit_hash: Optional[str] = None
    balena_uuid: Optional[str] = None
    balena_created_at: Optional[str] = None
    balena_revision: Optional[int] = None
    git_commit_hash: Optional[str] = None
    git_slug: Optional[str] = None


class CommitCreate(CommitBase):
    balena_id: int
    balena_commit_hash: str
    balena_uuid: str
    balena_created_at: str
    balena_revision: int


class CommitUpdate(CommitBase):
    """Git/Balena Commit Update schema."""

    pass


class CommitInDBBase(CommitBase):
    """Git/Balena Commit Database base schema."""

    id: int

    class Config:
        orm_mode = True


class Commit(CommitInDBBase):
    """Git/Balena Commit schema."""

    pass


class CommitInDB(CommitInDBBase):
    pass



"""Test Attachment Schema"""
from typing import Optional

from pydantic import BaseModel


class AttachmentBase(BaseModel):
    """Test Attachment Base schema."""

    name: Optional[str]
    filepath: Optional[str]


class AttachmentCreate(AttachmentBase):
    """Test Attachment Create schema."""

    name: str
    filepath: str


class AttachmentUpdate(AttachmentBase):
    """Test Attachment Update schema."""

    pass


class AttachmentInDBBase(AttachmentBase):
    """Test Attachment Database base schema."""

    id: int
    run_id: int

    class Config:
        orm_mode = True


class Attachment(AttachmentInDBBase):
    """Test Attachment schema."""

    pass


class AttachmentInDB(AttachmentInDBBase):
    pass



"""Models for messaging between the Test Executor and `pytest` processes."""

from typing import Dict, List, Optional, Union

from pydantic import BaseModel


class GUISettings(BaseModel):
    """Test Runner GUI Settings.

    Contains settings for configuring the Test Runner, including displaying
    serial number input dialogs, enabling a test interlock check, etc.
    """

    display_name: str = None
    stop_at_first_failure: bool = True
    autorun: bool = False
    dut_count: int = 1
    dut_grid: List[int] = [1, 1]
    pytest_args: str = None


class FixtureSettings(BaseModel):
    """Test Fixture Settings.

    Contains global settings specific to a given Fixture or Fixture Slot.

    - slot_id: Fixture Slot ID
    - api_url: Fixture Slot API URL (if applicable)
    - user_vars: User defined variables available for configuration via Test Runner
    """

    slot_id: str
    api_url: str = None
    user_vars: dict = {}


class TestCase(BaseModel):
    """Test Case configuration model.

    Contains information for the given test case. This is used to configure
    pass/fail limits and provide information that is displayed within the
    Test Runner GUI.

    The following information should be specified:

    - test_id: Test ID number used to identify the test case on the GUI
    - error_code: Error code to use within debugging guide
    - error_msg: Message to display in the GUI when a failure occurs for the
        given test case
    - description: Message displayed in the GUI for the test description
    - min_limit: Minimum limit, if applicable
    - max_limit: Maximum limit, if applicable
    - user_vars: User defined variables available for configuration via Test Runner
    """

    test_id: Optional[str] = None
    error_code: Optional[str] = None
    error_msg: Optional[str] = None
    description: Optional[str] = None
    min_limit: Optional[Union[str, int, float, bool]] = None
    max_limit: Optional[Union[str, int, float, bool]] = None
    user_vars: dict = {}


class TestPlanConfig(BaseModel):
    """Test Plan Configuration file model.

    Contains the complete test plan configuration information. This includes
    the test name, GUI configuration information, and any test case
    configuration information.
    """

    test_name: str
    test_cases: Dict[str, TestCase]
    runner_settings: GUISettings
    fixture_settings: Dict[str, FixtureSettings] = {"DEV": FixtureSettings(slot_id="1")}


class PytestTestPlanConfig(BaseModel):
    """Test Plan Configuration file model.

    Contains the complete test plan configuration information. This includes
    the test name, GUI configuration information, and any test case
    configuration information.
    """

    test_name: str
    test_cases: Dict[str, TestCase]
    runner_settings: GUISettings
    fixture_settings: FixtureSettings


class BalenaRelease(BaseModel):
    id: int
    commit: str
    created_at: str
    revision: int


class ConfigInfo(BaseModel):
    """Configuration Information Message

    Message structure for sending the current test plan revision information
    to the backend.
    """

    test_plan_version: str
    test_plan_path: str
    balena_uuid: Optional[str] = None
    balena_release: Optional[BalenaRelease] = None
    test_plan_config: TestPlanConfig



from .attachment import Attachment, AttachmentCreate, AttachmentInDB, AttachmentUpdate
from .commit import Commit, CommitCreate, CommitInDB, CommitUpdate
from .config import (
    BalenaRelease,
    ConfigInfo,
    FixtureSettings,
    GUISettings,
    PytestTestPlanConfig,
    TestCase,
    TestPlanConfig,
)
from .limits import Limits, LimitsCreate, LimitsInDB, LimitsUpdate
from .msg import Dialog, DialogResponse, Msg, Notif, SocketMsg, StatusBanner
from .plan import Plan, PlanCreate, PlanInDB, PlanUpdate
from .result import Result, ResultCreate, ResultInDB, ResultUpdate
from .run import Run, RunCreate, RunInDB, RunUpdate
from .token import Token, TokenPayload
from .user import User, UserCreate, UserInDB, UserUpdate



"""F3TS Plugin Utilities."""
import datetime
import logging
import pathlib
import ssl
import sys
from typing import Tuple, Union

import yaml
from websocket import create_connection

from . import schemas

logger = logging.getLogger(__name__)


def get_config_from_file(f3ts_config=None, slot_id=None):
    """Return a configuration object."""
    if f3ts_config:
        config_path = f3ts_config
    else:
        config_path = "config.yml"

    logger.info(f"config_path: {config_path}")
    with open(config_path) as f:
        config_dict = yaml.safe_load(f)
        config_obj = schemas.TestPlanConfig.parse_obj(config_dict)

    if slot_id:  # Parse into config for specified fixture:
        my_fixture_settings = schemas.FixtureSettings(slot_id=slot_id)
        for fixture_config in config_obj.fixture_settings.values():
            if slot_id == fixture_config.slot_id:
                my_fixture_settings = fixture_config

        f3ts_config_obj = schemas.PytestTestPlanConfig(
            test_name=config_obj.test_name,
            test_cases=config_obj.test_cases,
            runner_settings=config_obj.runner_settings,
            fixture_settings=my_fixture_settings,
        )
        logger.info(f"my test_plan_config: {f3ts_config_obj}")
        return f3ts_config_obj

    logger.info(f"my test_plan_config: {config_obj}")
    return config_obj


def get_settings_from_file(
    version: str, f3ts_config=None
) -> Tuple[schemas.LimitsCreate, schemas.GUISettings]:
    """Return particular settings for logging to database."""
    if f3ts_config:
        config_path = f3ts_config
    else:
        config_path = "config.yml"

    logger.info(f"config_path: {config_path}")
    with open(config_path) as f:
        config_dict = yaml.safe_load(f)
        runner_config_dict = config_dict.pop("runner_settings")
        config_dict["version"] = version
        limits_obj = schemas.LimitsCreate.parse_obj(config_dict)
        runner_config_obj = schemas.GUISettings.parse_obj(runner_config_dict)

    return limits_obj, runner_config_obj


def get_log_filename(
    test_name: str, sn: str, dirname: str, extension: str = "xml"
) -> str:
    """Return a log filename."""
    return str(
        pathlib.Path(dirname).joinpath(
            f"{test_name}"
            f"_{sn}"
            f"_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.{extension}"
        )
    )


def log_vars(record_property, recurse=1):
    """Log standard F3TS variables.

    Add the standard variables and expected configuration information to the
    JUnit-based XML test log. The following properties are recorded:

    - `test_id`: Test ID number
    - `error_code`: Error code for the test case
    - `error_msg`: Message to display when the test case fails
    - `min_limit`: Optional minimum limit
    - `max_limit`: Optional maximum limit
    - `description`: Optional test case description
    - `meas`: Value that is being measured and tested
    - `test_config`: F3TS fixture that provides test information from
       configuration files
        - The same values as listed above are logged.
    """
    caller = sys._getframe(recurse)
    if "test_config" in caller.f_locals.keys():
        for _k, _v in caller.f_locals["test_config"].dict().items():
            logger.info(f"PROPERTY: {_k} = {_v}")
            record_property(_k, _v)

    for k, v in caller.f_locals.items():
        if k in [
            "test_id",
            "error_code",
            "error_msg",
            "min_limit",
            "max_limit",
            "description",
            "meas",
        ]:
            logger.info(f"PROPERTY: {k} = {v}")
            record_property(k, v)


class UserDialog:
    """User Dialog Interface

    Object for sending user input prompts or popup messages either
    to the GUI or the serial terminal for pytest only configuration
    """

    def __init__(self, api):
        self.api = api
        self.websocket = None

        if self.api is None:
            print(
                "WARNING: Not connected to backend API, "
                "sending all user dialog prompts to terminal. "
                "Make sure pytest stdin is enabled using -s"
            )
        else:
            self.default_timeout = 2
            self.retries = self.api.retries
            self.api_url = self.api.api_url
            self.frontend_client_id = self.api.frontend_client_id
            self.backend_client_id = self.api.backend_client_id

            self.websocket_connect()

    def websocket_connect(self):
        """Setup websocket connection to frontend."""
        attempt = 0
        while True:
            try:
                websocket_url_head = self.api_url.replace("http://", "ws://").replace(
                    "https://", "wss://"
                )
                self.websocket_url = (
                    f"{websocket_url_head}/utils/runner_socket/"
                    f"{self.backend_client_id}"
                )

                # Define SSL options for the WebSocket connection
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                sslopt = {
                    "cert_reqs": ssl.CERT_NONE,
                    "check_hostname": False,
                    "ssl_context": ssl_context,
                }

                self.websocket = create_connection(
                    self.websocket_url, timeout=self.default_timeout, sslopt=sslopt
                )

                message = schemas.SocketMsg(
                    type="connect",
                    toClient=self.frontend_client_id,
                    fromClient=self.backend_client_id,
                )

                self.websocket.send(message.json())

                waiting = True
                while waiting:
                    response = self.websocket.recv()
                    response = schemas.SocketMsg.parse_raw(response)
                    if response.fromClient == self.frontend_client_id:
                        waiting = False

                return True

            except Exception:
                if attempt >= self.retries:
                    raise
                else:
                    attempt += 1

    def send_user_dialog(self, dialog: schemas.Dialog):
        """Send two-way user dialog message to the Backend."""

        self.websocket.settimeout(3600)

        message = schemas.SocketMsg(
            type="dialog",
            toClient=self.frontend_client_id,
            fromClient=self.backend_client_id,
            data=dialog,
        )

        # Generic error handling for any connection issues
        tries = 0
        while tries <= 1:
            try:
                self.websocket.send(message.json())

                # Wait until dialog response from connected frontent client
                waiting = True
                response = None
                while waiting:
                    response = self.websocket.recv()

                    response = schemas.SocketMsg.parse_raw(response)
                    if response.fromClient == self.frontend_client_id:
                        waiting = False

                break

            except:
                self.close()
                self.websocket_connect()
                self.websocket.settimeout(3600)

                tries += 1

        self.websocket.settimeout(self.default_timeout)
        if response:
            return schemas.DialogResponse.parse_obj(response.data)

    def send(
        self, message: schemas.Dialog, blocking: bool = True
    ) -> Union[schemas.DialogResponse, bool]:
        """Send Custom User Dialog to the GUI or terminal"""

        if self.api is None:
            response = self.send_terminal(message, blocking=blocking)

        else:
            if blocking:
                response = self.send_user_dialog(message)
            else:
                self.api.send_user_notif(message)
                return True

        return response

    def send_terminal(
        self, message: schemas.Dialog, blocking: bool = True
    ) -> schemas.DialogResponse:
        """Send Custom User Dialog to the pytest terminal

        Requires -s option to be enabled to allow stdin
        """

        response = schemas.DialogResponse()

        default_text = ""
        if message.defaultText:
            default_text = f"(ex/ {message.defaultText}\n"

        if message.inputType == "text":
            response.inputText = input(
                f"\n\n{message.title}\n{message.message}\n{default_text}--> "
            )
            response.okClicked = True
            response.cancelClicked = False

        elif blocking:
            print(f"\n\n{message.title}\n{message.message}\n{default_text}")

            ok = input(
                f"{message.okButtonText}? (y) " f"{message.cancelButtonText}? (n)\n--> "
            )
            if ok == "y":
                response.okClicked = True
                response.cancelClicked = False
            elif ok == "n":
                response.okClicked = False
                response.cancelClicked = True
            else:
                raise ValueError(f"{ok} is not a valid input.")

        else:
            print(f"\n\n{message.title}\n{message.message}\n{default_text}")

        return response

    def close(self):
        if self.websocket:
            self.websocket.close()


class StatusBanner:
    """Application Custom Status Interface

    Object for overriding the status banner in the application frontend
    """

    def __init__(self, api):
        self.api = api
        if self.api is None:
            print(
                "WARNING: Not connected to backend API, " "not updating status banner"
            )

    def override(self, override, status=None, color=None):
        if self.api is not None:
            if override:
                banner = schemas.StatusBanner(status=status, color=color)
            else:
                banner = None

            return self.api.send_banner_update(banner)

        elif override:
            print("STATUS: ", status)

        return False


class SerialNumber:
    """
    Object for setting and updating the status number in the
    application frontend
    """

    def __init__(self, api):
        self.api = api
        self.serial_number = ""
        if self.api is None:
            print(
                "WARNING: Not connected to backend API, " "not updating status banner"
            )

    def set(self, serial):
        self.serial_number = serial

        if self.api is not None:
            return self.api.send_serial_number(self.serial_number)

        return False



"""FixturFab Functional Test System `pytest` Plugin."""

from .fixtures import (
    auto_logger,
    backend_api,
    f3ts_assert,
    fixture_config,
    local_api,
    serial_number,
    status_banner,
    test_config,
    test_plan_config,
    user_dialog,
)
from .plugin import pytest_addoption, pytest_configure

__all__ = [
    # Fixtures
    test_config,
    test_plan_config,
    fixture_config,
    backend_api,
    local_api,
    auto_logger,
    f3ts_assert,
    user_dialog,
    status_banner,
    serial_number,
    # Hooks
    pytest_addoption,
    pytest_configure,
]



"""Test Case Schema"""
from datetime import datetime
from typing import Optional

from pydantic import BaseModel


class CaseBase(BaseModel):
    """Test Case Base schema."""

    name: Optional[str] = None
    config: Optional[str] = None
    user_properties: Optional[str] = None
    duration: Optional[int] = None
    result: Optional[bool] = None


class CaseCreate(CaseBase):
    """Test Case Create schema."""

    name: str
    config: str
    user_properties: str
    start: str = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")


class CaseUpdate(CaseBase):
    """Test Case Update schema."""

    result: bool
    pass


class CaseInDBBase(CaseBase):
    """Test Case Database base schema."""

    name: str
    run_id: int

    class Config:
        orm_mode = True


class Case(CaseInDBBase):
    """Test Case schema."""

    pass


class CaseInDB(CaseInDBBase):
    pass



"""F3TS Pytest Plugin Fixtures.

Various utility fixtures for accessing test plan configuration and test runner
parameters.
"""
import logging
import time

import pytest

from pytest_f3ts import schemas
from pytest_f3ts.runner_client import F3TSASCIIAPI, F3TSBackendAPI
from pytest_f3ts.utils import (
    SerialNumber,
    StatusBanner,
    UserDialog,
    get_config_from_file,
)

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def backend_api(pytestconfig) -> F3TSBackendAPI:
    """F3TS Backend Webservice API.

    Object for creating and sending requests to the Test Executer (TE)
    """
    try:
        return F3TSBackendAPI(
            api_url=pytestconfig.option.backend_url,
            run_id=pytestconfig.option.run_id,
            plan_id=pytestconfig.option.plan_id,
            slot_id=pytestconfig.option.slot_id,
            frontend_client_id=pytestconfig.option.frontend_client_id,
        )

    except Exception as err:
        print(f"WARNING: Unable to connect to backend API: {err}")


@pytest.fixture(scope="session")
def local_api(pytestconfig) -> F3TSASCIIAPI:
    """F3TS Backend Webservice API.

    Object for creating and sending requests to the Test Executer (TE)
    """
    return F3TSASCIIAPI(slot_id=pytestconfig.option.slot_id)


@pytest.fixture(scope="session")
def test_plan_config(pytestconfig, backend_api) -> schemas.PytestTestPlanConfig:
    """Fixture that provides access to the test plan configuration information.

    Returns the test plan configuration information. This is loaded from a YAML
    file that is titled either `config.yaml` within the same directory as the
    current test file, or can be defined by adding the `--f3ts_config` variable
    to the `pytest` command line call. For example:

    ```
    pytest -p f3ts --f3ts_config=path/to/config.yml
    ```

    The test plan must match the schema defined by the `TestPlanConfig` model.
    """
    if backend_api:
        config_obj = backend_api.get_config(
            slot_id=pytestconfig.option.slot_id,
        )
    else:
        config_obj = get_config_from_file(
            f3ts_config=pytestconfig.option.f3ts_config,
            slot_id=pytestconfig.option.slot_id,
        )

    logger.info(f"test_plan_config: {config_obj}")
    return config_obj


@pytest.fixture
def test_config(
    test_plan_config: schemas.PytestTestPlanConfig, request
) -> schemas.TestCase:
    """Fixture that provides test case configuration information.

    Returns the test case configuration information for the current
    test case. This requires the test case name to match the key within
    the configuration file, for example, a test named `test_5v0_voltage`
    would require the following section in the test configuration file:

    ```
    test_cases:
      test_5v0_voltage:
        test_id: 2.1
        error_code: 201
        error_msg: "Voltage out of spec"
        min_limit: 4.9
        max_limit: 5.1
        user_vars:
          user_dict_key1: user_dict_value1
          user_dict_key2: user_dict_value2
    ```
    """
    test_func_name = request.node.name.split("[")[0]

    if test_func_name in test_plan_config.test_cases.keys():
        return test_plan_config.test_cases[test_func_name]
    else:
        return None


@pytest.fixture(scope="session")
def fixture_config(
    test_plan_config: schemas.PytestTestPlanConfig, request, pytestconfig
) -> schemas.FixtureSettings:
    """Fixture that provides test fixture configuration information.

    Returns the test fixture configuration information for the current
    test fixture. This requires the test fixture id to match the fixture_id
    set the configuration file

    Required "pytest --slot_id" parameter specified on startup and
    within config.yml
    Required "pytest --fixture_id" parameter specified on startup and
    within config.yml:

    Example Use Case:

    ```yaml
    slot_1:
        slot_id: 1
        config:
            user_dict_key1: user_dict_value1
            user_dict_key2: user_dict_value2

    slot_2:
        slot_id: 2
        config:
            user_dict_key1: user_dict_value3
            user_dict_key2: user_dict_value4

    ```

    ```python
        def test_fixture_parameter(fixture_config):
            config_value = fixture_config.config['user_dict_key1']

            assert user_dict_value1==config_value, "Matches Fixture 1"
            assert user_dict_value3==config_value, "Matches Fixture 2"

    ```

    Start test program with --slot_id specified below:

        pytest --slot_id=1

    This grants the test runner access to slot_1's user defined parameters.
    This is useful for hardware serial numbers or MAC IDs that are different
    fixture to fixture but are required for the automated test software to
    function as desired.

    """
    if pytestconfig.option.slot_id is None:
        raise AssertionError(
            "ERROR: No Slot ID specified via --slot_id, "
            "no fixture_settings available"
        )

    return test_plan_config.fixture_settings


@pytest.fixture(autouse=True)
def auto_logger(test_config, record_property):
    """Log standard F3TS variables."""
    if test_config:
        record_property("test_config", dict(test_config))


@pytest.fixture
def f3ts_assert(backend_api, local_api, test_config, record_property, request):
    """Fixture to manually send assertion result to the Test Executer (TE).

    Assert statements made with f3ts_assert will show up as sub results within
    the test application GUI and stored as a result in the results database.
    """
    assert_id = 1
    start = time.time()

    def custom_assert(
        assert_value: bool = None,
        min_limit=None,
        meas=None,
        max_limit=None,
        test_id=None,
        test_name=None,
        description=None,
        error_code=None,
        error_msg=None,
    ):
        nonlocal assert_id, start
        args = locals()

        terminalreporter = request.config.pluginmanager.getplugin("terminalreporter")

        if backend_api:
            api = backend_api
        else:
            api = local_api

        assertion = schemas.ResultCreate(
            test_name=request.node.name,
            operator_id=api.operator_id,
            start=start,
            stop=time.time(),
            duration=time.time() - start,
            passed=bool(assert_value),
        )

        search_props = [
            "test_id",
            "description",
            "error_code",
            "error_msg",
            "min_limit",
            "max_limit",
            "meas",
        ]
        for key, value in args.items():
            if key in search_props:
                if args[key] is not None:
                    setattr(assertion, key, value)
                elif test_config and key in dict(test_config).keys():
                    setattr(assertion, key, dict(test_config)[key])

        assertion.test_id = str(assertion.test_id) + "." + str(assert_id)
        if assertion.meas is None:
            assertion.meas = assertion.passed

        assert_id += 1
        start = time.time()

        terminalreporter.ensure_newline()
        terminalreporter.section(f"Test ID: {assertion.test_id}", sep="-")

        terminalreporter.line(str(assertion))
        api.send_test_result(assertion)

        if not assertion.passed:
            print(" sending error notification to GUI ".center(80, "-"))

            error_message = "Subtest Failed"
            error_code = ""

            if "error_msg" in args.keys():
                error_message = args["error_msg"]
            elif "error_msg" in dict(test_config).keys():
                error_message = test_config.error_msg

            if "error_code" in args.keys():
                error_code = args["error_code"]
            elif "error_code" in dict(test_config).keys():
                error_code = test_config.error_code

            error_notif = schemas.Notif(
                title=f"ERROR {error_code}: {assertion.test_name}",
                message=error_message,
                okButtonText="OK",
            )

            terminalreporter.ensure_newline()
            terminalreporter.section(f"Error ID: {error_code}", sep="-")
            terminalreporter.line(str(error_notif))

            api.send_user_notif(error_notif)

        # Record test properties
        record_property("f3ts-assert", assertion.dict())
        assert assert_value

    yield custom_assert


@pytest.fixture(scope="session", autouse=True)
def user_dialog(backend_api) -> UserDialog:
    """Fixture that provides user dialog features.

    Sending dialog messages to terminal that requires
    user input require stdin enabled via:

        pytest -s

    """
    user_dialog = UserDialog(backend_api)

    yield user_dialog

    user_dialog.close()


@pytest.fixture(scope="session")
def status_banner(backend_api) -> StatusBanner:
    """Fixture that provides ability to override status banner."""
    return StatusBanner(backend_api)


@pytest.fixture(scope="session")
def serial_number(backend_api) -> SerialNumber:
    """Fixture for accessing the serial number of the device under test (DUT)."""
    yield SerialNumber(backend_api)


@pytest.fixture(scope="session")
def logfile_storage(backend_api):
    """Fixture to manually upload log files to the Test Executor (TE)."""

    def upload(filename: str, filepath: str):
        if backend_api is not None:
            attachment = schemas.AttachmentCreate(name=filename, filepath=filepath)

            backend_api.upload_logfile(backend_api.run_id, attachment)

    yield upload



"""Test Result Schema"""
from typing import Optional, Union

from pydantic import BaseModel


class ResultBase(BaseModel):
    """Test Result Message.

    Message structure for sending the results of a `pytest` test result back
    to the Test Executor.
    """

    test_id: Optional[str] = None
    error_code: Optional[str] = None
    error_msg: Optional[str] = None
    test_name: Optional[str] = None
    description: Optional[str] = None
    operator_id: Optional[int] = None
    min_limit: Optional[Union[str, int, float, bool]] = None
    max_limit: Optional[Union[str, int, float, bool]] = None
    meas: Optional[Union[str, int, float, bool]] = None
    start: Optional[float] = None
    stop: Optional[float] = None
    duration: Optional[float] = None
    passed: Optional[bool] = None


class ResultCreate(ResultBase):
    """Test Result Create schema."""

    test_id: Optional[str]
    test_name: str
    operator_id: int
    start: Optional[float]
    stop: Optional[float]
    duration: Optional[float]
    passed: bool


class ResultUpdate(ResultBase):
    """Test Result Update schema."""

    passed: bool
    pass


class ResultInDBBase(ResultBase):
    """Test Result Database base schema."""

    id: int
    uuid: str
    run_id: Optional[int] = None
    operator_id: int

    class Config:
        orm_mode = True


class Result(ResultInDBBase):
    """Test Result schema."""

    pass


class ResultInDB(ResultInDBBase):
    pass



"""Test Plan Schema"""
from typing import Dict, Optional

from pydantic import BaseModel

from .commit import Commit
from .config import FixtureSettings, GUISettings
from .limits import LimitsBase


class PlanBase(BaseModel):
    """Test Plan Base schema."""

    name: Optional[str] = None
    customer_name: Optional[str] = None
    project_name: Optional[str] = None
    part_number: Optional[str] = None
    version: Optional[str] = None
    path: Optional[str] = None
    limits_settings: Optional[LimitsBase] = None
    runner_settings: Optional[GUISettings] = None
    fixture_settings: Optional[Dict[str, FixtureSettings]] = None


class PlanCreate(PlanBase):
    """Test Plan Create schema."""

    path: str = "."


class PlanUpdate(PlanBase):
    """Test Plan Update schema."""

    pass


class PlanInDBBase(PlanBase):
    """Test Plan Database base schema."""

    id: int
    uuid: str
    owner_id: int
    commit_id: int = None
    commit: Optional[Commit] = None

    class Config:
        orm_mode = True


class Plan(PlanInDBBase):
    """Test Plan schema."""

    pass


class PlanInDB(PlanInDBBase):
    pass



"""Models for messaging between the Test Executor and `pytest` processes."""

from typing import Dict, Optional

from pydantic import BaseModel

from .config import TestCase


class LimitsBase(BaseModel):
    """Test Limits Base Model."""

    test_name: Optional[str] = None
    version: Optional[str] = None
    test_cases: Optional[Dict[str, TestCase]] = None
    # runner_settings: Optional[GUISettings] = None


class LimitsCreate(LimitsBase):
    """Test Limits Create"""

    test_name: str
    version: str
    test_cases: Dict[str, TestCase]
    # runner_settings: GUISettings


class LimitsUpdate(LimitsBase):
    """Git/Balena Limits Update schema."""

    pass


class LimitsInDBBase(LimitsBase):
    """Git/Balena Limits Database base schema."""

    id: int

    class Config:
        orm_mode = True


class Limits(LimitsInDBBase):
    """Git/Balena Limits schema."""

    pass


class LimitsInDB(LimitsInDBBase):
    pass
